import contextlib
import docker
import json
import os
import re
import shutil
import signal
import socket
import subprocess
import sys

from datetime import datetime, timezone
from tempfile import TemporaryDirectory

from app.utils import op_dev_service, OPLogFormatter, parse_log_msg


# There are two ways to run an op-dev service:
# 1. Using flask directly (specify the '--soft' flag)
#    - Much simpler
#    - Easier to test local library packages in this context (although also dangerous by the same token)
# 2. Wrapping in a docker container (default)
#    - More complex (dependency on docker)
#    - Can run exactly as it is run when deployed to a URL (no potential issues with requirements resolution, wsgi setup)
#    - Takes slightly longer

# Here are some useful commands if you run into issues with lingering containers:
# docker stop $(docker ps -a -q)
# docker rm $(docker ps -a -q)

# Here's a one-liner to kill any process running on port 5000:
# lsof -t -i tcp:5000 | xargs kill


RUN_SOFT_SCRIPT = '''
import asyncio
import json
import logging
import uvicorn

from op_service.logging_config import configure_logging
from op_service.op_version_metadata import op_version_metadata

async def main():
  configure_logging()
  from app import service
  logging.info('Initialized process: ' + json.dumps(op_version_metadata))
  return service.starlette

app = asyncio.run(main())


# TODO: I don't know what gives, maybe I have a local dependency installed that's out of date or some shit, but any
# communication via SSL inside the asyncio loop seems to fail without the following:
import ssl
ssl._create_default_https_context = ssl._create_unverified_context


# TODO: try this
# It's possible that the service started some async tasks when it was imported
# loop = asyncio.get_event_loop()
# loop.run_until_complete(asyncio.gather(*asyncio.all_tasks()))

# uvicorn app:starlette --host 0.0.0.0 --port 5000 --workers 1 --no-access-log

uvicorn.run(app, host='0.0.0.0', port=%(port)s, access_log=False)
'''


# This should NOT be kept separately - it is duplicated in the op-dev-manager-service package's build code for services
DEFAULT_DOCKERFILE = '''
FROM opdevio/op-dev-service

# Install the requirements first to take advantage of caching
COPY requirements.txt /var/www/
RUN pip install -r /var/www/requirements.txt

# Only copy over the 'app' directory
COPY app /var/www/app

COPY config.json /var/www/.op-dev/config.json
COPY metadata.json /var/www/.op-dev/metadata.json
'''


def _docker_build(build_context_dir, no_cache=False):
  build_output = ''
  try:
    docker_client = docker.from_env()
    stream = docker_client.api.build(path=build_context_dir, rm=True, nocache=no_cache)
    image_id = None
    for chunk in stream:
      for line in chunk.splitlines():
        status = json.loads(line)
        if 'error' in status:
          raise Exception('Error building docker image: %s' % status['error'])
        if 'stream' in status:
          msg = status['stream']
          build_output += msg

          # Note: this seems incredibly fragile, but it appears to be the best way to get the image_id...
          # https://github.com/docker/docker-py/blob/39bb78ac694d8d6e53882d3dbc9ebc3c92f5519d/docker/models/images.py#L176
          match = re.search(r'(^Successfully built |sha256:)([0-9a-f]+)$', msg)
          if match:
            image_id = match.group(2)
  except:
    # Only print the build output if we failed - should be so fast most times it's not worth flashing over the screen.
    # It also won't show up in production, so let's try to mimic that environment as much as possible.
    print(' \033[31mFAILED\033[0m')
    build_output and print(build_output)
    raise

  if not image_id:
    raise Exception('Error building docker image: could not determine image ID from build output')

  return image_id


def _shutdown(container):
  print('\033[36mShutting down...\033[0m')
  container.kill()
  container.remove(force=True)
  sys.exit()


def _check_port(port):
  with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
    if s.connect_ex(('localhost', port)) == 0:
      # Port is already in use
      docker_client = docker.from_env()
      for container in docker_client.containers.list():
        for port_list in container.ports.values():
          if str(port) in [ x.get('HostPort') for x in (port_list or []) ]:
            cmd = 'docker kill %s' % container.id
            msg = 'A different docker container is already running on port %i. Try a different port using --port PORT, or kill that container using this command: %s' % (port, cmd)
            sys.exit(msg)
      sys.exit('Port %i is already in use. Try a different port using --port PORT.' % port)


def _get_repository_url(pkg_dir):
  # Returns None if the URL cannot be determined (this is not a git package or it has some weird configuration)
  cmd = [ 'git', 'config', '--local', '--get', 'remote.origin.url' ]
  repository_url = subprocess.run(cmd, capture_output=True, cwd=pkg_dir).stdout.decode('utf-8').strip()
  return repository_url or None


# Returns the config ID, or none if not applicable
# Postconditions: the file at 'dest_config_path' contains the config data as specified
def _create_config(config_id, config_path, config_namespace, pkg_dir, dest_config_path):
  if config_path:
    shutil.copyfile(config_path, dest_config_path)
    return None

  config = None
  if config_id:
    response = op_dev_service.call('get-config', { 'config_id': config_id })
    config = response['value']
  else:
    repository_url = _get_repository_url(pkg_dir)
    if repository_url:
      params = { 'repository_url': repository_url, 'config_namespace': config_namespace }
      # TODO: this is overly general - we should catch a specific exception (namespace not found) and throw on others
      with contextlib.suppress(Exception):
        response = op_dev_service.call('get-latest-config', params)
        config = response['config']['value']
        config_id = response['config_id']
  
  if not config:
    print("\rNo config found for namespace '%s' (using empty)" % config_namespace, file=sys.stderr)
    config = '{}'

  with open(dest_config_path, 'w') as file:
    file.write(config)

  return config_id


def _create_metadata(path, port, pkg_id, commit_id, config_id):
  metadata = {
    'domain': 'http://localhost:%i' % port,
    'pkg_id': pkg_id,
    'commit_id': commit_id,
    'config_id': config_id
  }
  content = json.dumps(metadata)
  with open(path, 'w') as metadata_file:
    metadata_file.write(content)


def _is_web_client(pkg_dir):
  package_json = os.path.join(pkg_dir, 'package.json')
  return os.path.exists(package_json)


def run(parser, args):
  # Make sure our desired port is not currently in use
  port = int(args.port) if args.port else (8080 if _is_web_client(args.pkg_dir) else 5000)
  _check_port(port)

  with TemporaryDirectory() as tmp_dir:

    print('\033[36mGetting configuration...\033[0m', end='', flush=True)
    config_path = os.path.join(tmp_dir, 'config.json')
    config_id = _create_config(args.config_id, args.config_path, args.config_namespace, args.pkg_dir, config_path)

    if _is_web_client(args.pkg_dir):
      print('\033[2K\r', end='', flush=True)

      if args.main:
        # TODO: you should probably be able to run this way too
        sys.exit("'--main' option is only for service packages")

      env = os.environ.copy()
      subprocess.check_call([ 'npm', 'install' ], cwd=args.pkg_dir, env=env)

      if config_id:
        env['OP_DEV_CONFIG_ID'] = config_id
      else:
        env['OP_DEV_CONFIG_PATH'] = config_path

      # This call hangs until explicitly terminated (for hot-reloading)
      # Note that by default, this command would find the next open port (from 8080), but "explicit is better than
      # implicit"...
      subprocess.check_call([ 'npm', 'run', 'serve', '--', '--port', str(port) ], cwd=args.pkg_dir, env=env)

    # You previously only used the --soft argument to run a single instance of the code as a script
    # if args.soft:
    #   metadata_file_path = os.path.join(tmp_dir, 'metadata.json')
    #   _create_metadata(metadata_file_path, port, None, None, config_id)

    #   env = os.environ.copy()
    #   env['OP_DEV_CONFIG'] = config_path
    #   env['OP_DEV_METADATA'] = metadata_file_path
    #   env['OP_DEV_SERVICE_TEST'] = '1'
    #   env['PYTHONPATH'] = args.pkg_dir

    #   cmd = [ 'python3', '-u', '-c', RUN_SOFT_SCRIPT % { 'port': port } ]
    #   # Capture all output and pass it through our op-dev log parser
    #   process = subprocess.Popen(cmd, stderr=subprocess.STDOUT, stdout=subprocess.PIPE, env=env)
    #   stream = process.stdout

    else:
      build_context = os.path.join(tmp_dir, 'pkg')
      shutil.copytree(args.pkg_dir, build_context)
      shutil.move(config_path, build_context)

      metadata_file_path = os.path.join(build_context, 'metadata.json')
      _create_metadata(metadata_file_path, port, None, None, config_id)

      docker_file_path = os.path.join(build_context, 'Dockerfile')
      if not os.path.exists(docker_file_path):
        with open(docker_file_path, 'w') as file:
          file.write(DEFAULT_DOCKERFILE)

      volumes = {}
      if args.main is not None:
        volumes[ os.getcwd() ] = { 'bind': '/main', 'mode': 'rw' }
        with open(docker_file_path, 'a') as file:
          # When running the __main__ file, I don't want developers to have to configure the op_service logging, it
          # should just work. This is definitely some magic, and I don't like this pattern, but it does work...
          file.write('RUN echo "import op_service.logging_config\\nop_service.logging_config.configure_logging()\\n$(cat /var/www/app/__main__.py)" > /var/www/app/__main__.py\n')
          file.write('WORKDIR /main\n')

      print('\033[2K\r\033[36mBuilding...\033[0m', end='', flush=True)
      image_id = _docker_build(build_context, no_cache=args.no_cache)
      print('\033[2K\r', end='', flush=True)

      # Now we can run the container
      docker_client = docker.from_env()
      container = docker_client.containers.run(
        image_id,
        command=None if args.main is None else [ 'python', '-B', '-m', 'app', *args.main ],

        # TODO: we should really use the package name here
        name=os.path.basename(os.path.abspath(args.pkg_dir)),

        detach=True,
        remove=True,
        ports={ 5000: port },
        environment={ 'OP_DEV_SERVICE_TEST': '1' },
        volumes=volumes
      )
      stream = container.logs(stream=True)

      # Register signal handlers so we can clean up the container correctly
      signal.signal(signal.SIGINT, lambda *args: _shutdown(container))
      signal.signal(signal.SIGTERM, lambda *args: _shutdown(container))
      signal.signal(signal.SIGPIPE, lambda *args: _shutdown(container))

    # Apply our log formatter to the stream output
    formatter = OPLogFormatter(time_format='%-I:%M:%S %p')
    line = ''
    for encoded_chunk in stream:
      # The stream can deliver partial lines, but we do rely on the behavior that a raw newline character cannot be
      # contained within a chunk.
      chunk = encoded_chunk.decode('utf-8')

      # There is an issue here where access logs can get interspersed in the middle of large chunks. We rely on the fact
      # that these access log chunks are small, and thus likely to be contained within one chunk. This is pretty fragile -
      # a better solution would be to split up the logging to avoid writing to just one stream. We may have issues if
      # multiple workers are generating lots of large log messages at once as well, which this workaround will not solve.
      # Not too big an issue, as the only consequence is poorly-formatted log messages.
      # if chunk.startswith('[OP-ACCESS]'):
      #   print(formatter.format(datetime.now(timezone.utc), parse_log_msg(chunk[:-1])))
      #   continue

      line += chunk
      if line.endswith('\n'):
        print(formatter.format(datetime.now(timezone.utc), parse_log_msg(line[:-1])))
        line = ''



from app.utils import op_dev_service


def run(parser, args):
  if args.operation != 'get':
    return

  credentials = op_dev_service.call('get-code-commit-credentials-for-team', { 'team_id': args.team_id })
  print('username=%s' % credentials['user_name'])
  print('password=%s' % credentials['password'])



# https://dev.to/aschmelyun/using-the-magic-of-mkcert-to-enable-valid-https-on-local-dev-sites-3a3c


import os
import shlex
import socket
import subprocess


OP_DEV_HOME = os.path.expanduser('~/.op-dev')
CERT_PATH = os.path.join(OP_DEV_HOME, 'localhost+4.pem')
KEY_PATH = os.path.join(OP_DEV_HOME, 'localhost+4-key.pem')


def verify():
  p = subprocess.run([ 'ipconfig', 'getifaddr', 'en0' ], capture_output=True, check=True)
  ip_addr = p.stdout.decode('UTF-8').strip()

  local_network_addr = socket.gethostname() + '.local'

  os.makedirs(OP_DEV_HOME, exist_ok=True)
  subprocess.run([ 'mkcert', '-install' ], check=True, cwd=OP_DEV_HOME)
  subprocess.run([ 'mkcert', 'localhost', '127.0.0.1', ip_addr, local_network_addr, '::1' ], check=True, cwd=OP_DEV_HOME)


def run(parser, args):
  verify()
  print('KEY=' + shlex.quote(KEY_PATH) + ' CERT=' + shlex.quote(shlex.quote(CERT_PATH)))



import contextlib
import json

from op_service_client import OPServiceClient

from app.config import get_config


config = get_config()
url = config.get('op_dev_api_url', 'https://op-dev-manager-service.op-dev.io')
auth = { 'route': 'get-jwt-from-secret', 'params': config }
op_dev_service = OPServiceClient(url, auth)


# TODO: some of this logic is still obsolete with the switch from nginx / flask to gunicorn / starlette...

# Command-line color codes:
# https://www.lihaoyi.com/post/BuildyourownCommandLinewithANSIescapecodes.html


DEFAULT_LOG_LEVEL_STYLES = {
  'TRACE': ("\033[90m", "\033[0m"),
  'DEBUG': ("\033[32m", "\033[0m"),
  'WARNING': ("\033[33m", "\033[0m"),
  'ERROR': ("\033[31m", "\033[0m"),
  'CRITICAL': ("\033[91m", "\033[0m")
}


DEFAULT_LOG_NAME_STYLES = {
  'OP-DEV': ("\033[90m", "\033[0m"),
  'OP-TEST': ("\033[1m", "\033[0m"),
  'op-request': ("\033[36m", "\033[0m")
}


def parse_log_msg(msg):
  '''
  Log messaegs produced using the op-dev pattern should be a string prefixed with '[OP]' and followed by a JSON object
  that should contain the following keys:
    l: the level with which the message was logged - DEBUG, INFO, WARNING, ERROR, CRITICAL, or None
    n: the name of the logger, as in logging.getLogger(<name>) - will be 'root' instead of None
    i: the instance ID of the service from which this message was logged
    s: the python source file path (relative to the module root)
    #: the line number from the above source file
    m: the message string (potentially contains newlines)

  Valid OP log messages will return the dictionary containing those keys. Other messages will only have the 'm'
  property set to the 'msg' as passed in.

  Note: CloudWatch may incorrectly conflate log messages that involve characters like '\r' (or possibly other
  characters with special meaning) that could muck with our log lines. The impact is that our formatted log messages
  may not get parsed correctly if our tag follows one of these characters. Thus, if the '\r' character precedes a known
  prefix, we ignore the content up to and including that character. This makes sense, as that output would have been
  overwritten in the terminal anyway, even though it contains some information. OP log chunks always end with a
  newline, so there is no need to check for '\r' after a prefix token is found.
  '''
  for i, char in enumerate(msg):
    if i == 0 or char == '\r':
      start_at_i = 0 if i == 0 else i + 1
      if msg[start_at_i:(start_at_i + 4)] == '[OP]':
        with contextlib.suppress(Exception):
          return json.loads(msg[start_at_i + 4:])
  return { 'm': msg }


class OPLogFormatter(object):

  def __init__(self, lvl_styles=DEFAULT_LOG_LEVEL_STYLES, name_styles=DEFAULT_LOG_NAME_STYLES, tz=None, time_format=None):
    self.lvl_styles = lvl_styles
    self.name_styles = name_styles
    self.tz = tz
    self.time_format = time_format

  def _is_op_msg(self, msg_dict):
    return 'i' in msg_dict

  def format_time(self, t):
    local_t = t.astimezone(self.tz)
    if self.time_format:
      return local_t.strftime(self.time_format)
    return local_t.isoformat()

  def format(self, t, parsed_msg):
    # Styles for log names override those for the log levels
    # TODO: consider making this additive rather than replacing other styles
    style = ('', '')
    style = self.lvl_styles.get(parsed_msg.get('l'), style)
    style = self.name_styles.get(parsed_msg.get('n'), style)
    style_prefix, style_suffix = style
    t_str = self.format_time(t)

    if self._is_op_msg(parsed_msg):
      # OP messages get a prefixed header row containing the metadata
      prefix = '[' + t_str + '] ' + ('[%(l)s] [%(n)s] [%(i)s] [%(s)s#%(#)s]\n' % parsed_msg)
    else:
      # Non-OP messages get the timestamp printed on every line (since we don't know the boundaries of the message)
      prefix = '[' + t_str + '] '

    if parsed_msg.get('n') == 'op-request':
      # Apply special formatting to op-request log messages
      try:
        request_data = json.loads(parsed_msg['m'])
        if request_data['_'] == 0:
          msg = 'Started request to route: %(r)s' % request_data
        else:
          msg = 'Finished request to route: %(r)s [%(c)s]' % request_data
          if 'pv' in request_data:
            msg += '\nInput profile (validated):\n%s' % json.dumps(request_data['pv'], indent=2)
          elif 'p' in request_data:
            msg += '\nInput profile:\n%s' % json.dumps(request_data['p'], indent=2)
          if 'iv' in request_data:
            msg += '\nInput data (validated):\n%s' % json.dumps(request_data['iv'], indent=2)
          elif 'i' in request_data:
            msg += '\nInput data:\n%s' % json.dumps(request_data['i'], indent=2)
          if 'o' in request_data:
            msg += '\nResponse data:\n%s' % json.dumps(request_data['o'], indent=2)
      except:
        # Only for legacy services
        msg = parsed_msg['m']
    else:
      msg = parsed_msg['m']

    return style_prefix + prefix + msg + style_suffix



import setuptools
import sys

assert sys.version_info >= (3, 6)

setuptools.setup(
  name='op-dev',
  version='1.0',
  author='Alex Exarhos',
  author_email='AlexExarhos@gmail.com',
  packages=[ 'app' ],
  entry_points={
    'console_scripts': [
      'op-dev = app.__main__:main'
    ]
  },
  install_requires=[
    'boto3 ~= 1.18.27',
    'op-service-client ~= 0.1.0',
    'python-dateutil ~= 2.8.1',
    'pytz ~= 2021.3',
    'docker ~= 4.1.0',
    'gitpython ~= 3.0.5',
    'tzlocal ~= 2.0.0'
  ]
)
